'''Python logging setup and database backend.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

This package provides basic support tools for the standard Python
:mod:`logging` package.  In particular, it exposes the
:mod:`logging.config` setup language through configuration layers
such as :mod:`yakonfig`, and it includes some basic database
support utilities.

Applications that use this support a ``logging`` block in their
configuration files.  This may have any valid logging setup supported
by the :mod:`logging.config` module.  The default configuration
includes a formatter named ``fixed`` and a handler named ``console``
bound to the root logger, and a further console handler named
``debug`` not bound anywhere.

Command-line options
====================

.. program:: dblogger

All programs that use :mod:`dblogger` support shared command-line
options.  If not changed in the configuration, the defaults write
:data:`logging.INFO` level output to the console, and
:option:`--verbose` enables all debug-level logging.

.. option:: --verbose, -v

Include more output, increasing the ``console`` log handler's level
one step.  This can be included multiple times.  If the program is
more verbose than quiet and the ``console`` log handler is not
attached to the root logger, adds it.

.. option:: --quiet, -q

Include less output, decreasing the ``console`` log handler's level
one step.  This can be included multiple times.

.. option:: --debug <logger>

Write debug-level output for `logger` to the ``debug`` log handler.
This can be included multiple times.

Recipes
=======

To enable debug-level logging everywhere in an application using this
package alongside a YAML-based configuration system, include a
top-level block in the YAML file:

.. code-block:: yaml

    logging:
      root:
        level: DEBUG

To enable debug-level logging for a specific package in the
configuration file:

.. code-block:: yaml

    logging:
      loggers:
        some.other.package:
          level: DEBUG

To write debug logs globally to a file:

.. code-block:: yaml

    logging:
        file:
          class: logging.FileHandler
          filename: /tmp/dblogger.log
          formatter: fixed
      root:
        handlers: [console, file]

For further details about what is allowed, see the Python library
:mod:`logging.config` documentation.

Logging tools
=============

.. autoclass:: DatabaseLogHandler
   :show-inheritance:

.. autoclass:: FixedWidthFormatter
   :show-inheritance:

Initial logging setup
=====================

Include :mod:`dblogger` in your :func:`yakonfig.parse_args` call
to do all of the initial setup.  When control returns to your
application, logging will be fully set up.

Older programs that do not yet support this interface use:

.. autofunction:: configure_logging

:program:`dblogger` query tool
==============================

.. automodule:: dblogger.query

'''
from __future__ import absolute_import
from dblogger.configure import config_name, default_config, add_arguments, \
    runtime_keys, normalize_config
from dblogger.configure import configure_logging
from dblogger.format import FixedWidthFormatter
from dblogger.logger import DatabaseLogHandler
from dblogger.query import *



''':mod:`logging` handler that stores log messages in a database.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

'''
from __future__ import absolute_import

from importlib import import_module
import time
import logging
import json
import cPickle as pickle
import random
import struct
import sys

from uuid import UUID
from tblib import pickling_support
pickling_support.install()  ## register traceback smarts with pickle

from dblogger.utils import gen_uuid
import kvlayer
import yakonfig


class DatabaseLogHandler(logging.Handler):
    '''Log handler that stores log messages in a database.

    This uses :mod:`kvlayer` to store the actual log messages.
    When the log handler is created, the caller needs to pass
    in either the :mod:`kvlayer` configuration or the actual
    :class:`kvlayer.AbstractStorage` object.  The constructor
    also accepts a virtual table name, defaulting to ``log``.

    If a global YAML file is used to configure the application, then
    YAML reference syntax can be used to share this handler's
    configuration with the global kvlayer configuration.

    .. code-block:: yaml

        kvlayer: &kvlayer
          storage_type: redis
          storage_addresses: [ 'redis.example.com:6379' ]
        logging:
          handlers:
            db:
              class: dblogger.DatabaseLogHandler
              storage_config: *kvlayer

    Log messages are stored in a table with a single UUID key, where
    the high-order bits of the UUID are in order by time.  The actual
    table values are serialized JSON representations of the log
    records.

    This log handlers adds a format string property ``%(humantime)s``
    with a time in a fixed format, and ``%(exc_text)s`` with the
    formatted traceback from an exception.  These properties are also
    included in the JSON stored in the database..

    .. automethod:: __init__

    '''
    def __init__(self, storage_client=None, table_name="log",
                 storage_config=None):
        """Create a new database log handler.

        You must either pass in ``storage_client``, an actual kvlayer
        client object, or ``storage_config``, a dictionary which will
        be passed to ``kvlayer.client()``.  Log messages
        will be stored in the table ``table_name``.

        :param storage_client: existing storage client
        :type storage_client: :class:`kvlayer.AbstractStorage`
        :param str table_name: virtual table name
        :param dict storage_config: configuration for new storage client

        """
        super(DatabaseLogHandler, self).__init__()

        if storage_client is None:
            if storage_config is None:
                raise RuntimeError('must pass either storage_client or '
                                   'storage_config')
            with yakonfig.defaulted_config(
                    [kvlayer], config=dict(kvlayer=storage_config)):
                storage_client = kvlayer.client()
            
        self.storage = storage_client
        self.table_name = table_name
        storage_client.setup_namespace({ table_name : 1 })

    def formatDBTime(self, record):
        record.humantime = time.strftime('%Y-%m-%dT%H:%M:%S-%Z', time.localtime(record.created))

    @classmethod
    def serialize(cls, record):
        return pickle.dumps(record.__dict__,  protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def deserialize(cls, rec_pickle):
        xdict = pickle.loads(rec_pickle)
        return logging.makeLogRecord(xdict)

    def emit(self, record):
        '''
        handle a record by formatting parts of it, and pushing it into
        storage.
        '''
        self.format(record)
        self.formatDBTime(record)

        if record.exc_info:
            record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
        else:
            record.exc_text = ''

        new_uuid = gen_uuid(record.created)
        dbrec = self.serialize(record)
        self.storage.put(self.table_name, ((new_uuid,), dbrec))
